/*===============================================================*/

#include <stdio.h>
#include <assert.h>
#include "parti.h"

#define CHARMM

/*=================================================================
 * Function:   localize()
 *=================================================================*/

void  PARTI_flocalize(  tabptr, hash_table, global_refs,  ndata,
                        no_off_proc,  my_size, mask, new_stamp, new_mask)

TTABLE   *tabptr;
HASHSTR  *hash_table;
int      *global_refs;
int       ndata;
int      *no_off_proc;
int       my_size;
int       mask;
int      *new_stamp;
int      *new_mask;
{
  int      dont_care;
  int      stamp=0;
 
  assert( tabptr != NULL );
  assert( hash_table != NULL );

  stamp = (*new_stamp  == NO_STAMP ) ?  NO_STAMP : 0 ;

  PARTI_hash ( tabptr,  hash_table, mask, &stamp,
               global_refs, ndata, my_size, &dont_care, no_off_proc);

  if (*new_stamp != NO_STAMP ) {
      *new_stamp = stamp;
      *new_mask  = PARTI_make_mask( stamp );
  }

  return;
}

/*=================================================================
 * Function:   parti_flocalize()
 *=================================================================*/

void  parti_flocalize_(  tabindex, hashindex, global_refs,  ndata,
			no_off_proc,  my_size, mask, new_stamp, new_mask)
int  *tabindex, *hashindex, *global_refs, *ndata ;
int  *no_off_proc, *my_size, *mask, *new_stamp, *new_mask;
{

  TTABLE   *tabptr = NULL;
  HASHSTR  *hash_table;
  int       i;

  hash_table  =  (HASHSTR *)get_hashstr (*hashindex);
  tabptr      =  (TTABLE *)get_table_ptr(*tabindex);

  for(i=0;i<*ndata;i++) {
    if ( global_refs[i] < 0)
      global_refs[i]++;
    else 
      global_refs[i]--;
   }

  PARTI_flocalize( tabptr, hash_table, global_refs,  *ndata,
		   no_off_proc,  *my_size, *mask, new_stamp, new_mask);

  for(i=0;i<*ndata;i++) {
    if ( global_refs[i] < 0)
      global_refs[i]--;
    else 
      global_refs[i]++;
   }

  return ;
}

/*===================================================================*/

void  PARTI_rmlocalize( tabptr, hash_table, global_refs, ndata,
                       strip_size, my_size, old_mask,
	               new_stamp, new_mask, off_proc )

TTABLE  *tabptr;
HASHSTR *hash_table;
int     *global_refs; 
int      ndata;
int      strip_size, my_size, old_mask;
int     *new_stamp, *new_mask;
int     *off_proc;
{
  /*...................................................*/
  int no_iter, max_iter, dont_care, extra;
  int offset,  place, offprocessor, *temp_array;
  int array_size, mask, combo_mask, stamp;
  int nstamp, nmask,  offp, noffset;
  int i;
  void getcopy(), putcopy(); 
  /*...................................................*/

  no_iter  = ndata / strip_size ;
  max_iter = no_iter;
  PARTI_gihigh( &max_iter, 1, &dont_care );
  extra    = ndata - no_iter * strip_size;

  noffset      = offset =  my_size;
  place        = 0;
  offprocessor = offp = 0;
  temp_array = (int *) malloc ( sizeof(int) * strip_size );

  /*...................................................*/
  for ( i=0; i<max_iter; i++ ) {

    array_size  = ( i >= no_iter ) ? 0 : strip_size ;
    mask        = ( i == 0 ) ? old_mask : combo_mask;
    stamp       = ( i == 0 ) ?  0 : NO_STAMP;
    getcopy ( global_refs,  place, temp_array, array_size);

    PARTI_hash( tabptr, hash_table, mask, &stamp ,
		temp_array, array_size, noffset,  &offp, &dont_care); 
   
    putcopy ( global_refs,  place, temp_array, array_size);

    if ( i == 0 ) {
       nstamp  =  stamp;
       nmask   =  PARTI_make_mask( stamp );
       combo_mask = ( old_mask != NO_MASK ) ?
                    ( old_mask | nmask )    :  nmask;
    } 

    if ( i <= no_iter ) {
      place        += array_size;
      offprocessor  = offp ;
      noffset       = offset + offp ;
    }
  }
  /*...................................................*/

  assert ( extra >= 0 );
  mask        = ( i == 0 ) ? old_mask : combo_mask;
  stamp       = ( i == 0 ) ?  0 : NO_STAMP;
  getcopy( global_refs,  place, temp_array, extra );

  PARTI_hash ( tabptr, hash_table, mask, &stamp, temp_array, extra,
               noffset, &offp , &dont_care);		

  if ( i == 0 ) {
       nstamp  =  stamp;
       nmask   =  PARTI_make_mask( stamp );
    }

  putcopy( global_refs, place, temp_array, extra );

  noffset       = offset + offp  ;
  offprocessor  = offp  ;
  /*...................................................*/

  *new_stamp = nstamp;
  *new_mask  = (nmask|old_mask);
  *off_proc  = offprocessor;
  free( temp_array);
  return;
}  

/*================================================================*/

void  getcopy( global_refs,  place, temp_array, array_size)
int *global_refs, *temp_array;
int  place, array_size;
{
 int j;
 int x;

 for ( j=0; j<array_size; j++ ) {
           x  = global_refs[place+j];
#          ifdef CHARMM
           x  = ( x > 0 ) ? x : (-x);
#          endif
           temp_array[j] = x; 
 }
}
/*================================================================*/
void  putcopy( global_refs,  place, temp_array, array_size)
int *global_refs, *temp_array;
int  place, array_size;
{
 int j ;
 int x,y;
 
 for ( j=0; j<array_size; j++ ) {
           x  = global_refs[place+j];
           y  = temp_array[j];
#          ifdef CHARMM
           y  = ( x >= 0 ) ? y : (-y);
#          endif
           global_refs[place+j] = y;
  }
}
/*================================================================*/

void  parti_rmlocalize_(  tabindex, hashindex, global_refs,  ndata,
                          strip_size, my_size, old_mask,
			  new_stamp, new_mask, off_proc )

int  *tabindex,   *hashindex, *global_refs, *ndata ;
int  *strip_size, *my_size, *old_mask;
int  *new_stamp,  *new_mask, *off_proc;
{
  TTABLE   *tabptr = NULL;
  HASHSTR  *hash_table;
  int       i;

  hash_table  =  (HASHSTR *)get_hashstr (*hashindex);
  tabptr      =  (TTABLE *)get_table_ptr(*tabindex);
  for(i=0;i<*ndata;i++) {
    if ( global_refs[i] > 0 ) 
      global_refs[i]--;
    else if ( global_refs[i] < 0)
      global_refs[i]++;
   }

  PARTI_rmlocalize( tabptr, hash_table, global_refs, *ndata,
	            *strip_size, *my_size, *old_mask,
	            new_stamp, new_mask, off_proc );

  for(i=0;i<*ndata;i++) {
     if ( global_refs[i] > 0 )
       global_refs[i]++;
     else if ( global_refs[i] < 0 )
       global_refs[i]--;
  }

  return ;
}

/*================================================================*/
